# -*- coding: utf-8 -*-
# Copyright (c) 2025, Alexei Znamensky
# All rights reserved.
#
# This file is part of the Andebox project and is distributed under the terms
# of the BSD 3-Clause License. See LICENSE file for details.
from andebox.cli import run

if __name__ == "__main__":
    run()

# code: language=python tabSize=4



# -*- coding: utf-8 -*-
# Copyright (c) 2024-2025, Alexei Znamensky
# All rights reserved.
#
# This file is part of the Andebox project and is distributed under the terms
# of the BSD 3-Clause License. See LICENSE file for details.
import hashlib
import json
import re
from io import StringIO
from pathlib import Path
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

try:
    from ruamel.yaml import YAML

    IMPORT_ERROR = None
except ImportError as e:
    IMPORT_ERROR = e

from ..exceptions import AndeboxException
from .base import AndeboxAction


FIXME_TAG = "__FIXME__"
JSON_SAMPLE_PREFIX = "JSONSAMPLE"


def fixme(s):
    return f"{FIXME_TAG}({s})"


DESCRIPTION_ACCEPTED_END_CHARS = (".", "!", ":", ";", ",", "?")
OFFENDING_SPEC = [
    # processed in order, apply only the first hit
    # if using parenthesis in regexp, they MUST be non-capturing, as in (?:pattern1|pattern2|...)
    dict(regexp=r"`", apply=fixme),
    dict(regexp=r"\bi\.e\.?\b", apply=fixme),  # i.e
    dict(regexp=r"\be\.g\.?\b", apply=fixme),  # e.g.
    dict(regexp=r"[^/]etc\b", apply=fixme),  # etc, but not /etc
    dict(regexp=r"\bvia\b", apply=fixme),  # via
    dict(regexp=r"\bversus\b", apply=fixme),
    dict(regexp=r"\bvs\.?\b", apply=fixme),
    dict(regexp=r"\bversa\b", apply=fixme),
    dict(regexp=r"won't", replace="will not"),
    dict(regexp=r"\b(?:[Ww]i|')ll\b", apply=fixme),
    dict(regexp=r"[a-zI]'ve", apply=lambda s: s.replace("'ve", " have")),
    dict(regexp=r"can't", replace="cannot"),
    dict(regexp=r"Can't", replace="Cannot"),
    dict(regexp=r"[a-zI]n't", apply=lambda s: s.replace("n't", " not")),
    dict(regexp=r"[a-zI]'re", apply=lambda s: s.replace("'re", " are")),
    dict(regexp=r"[a-zI]'d", apply=lambda s: s.replace("'d", " would")),
    dict(regexp=r"let's", replace="let us"),
    dict(
        regexp=r"(?:there|s?he|it)'s (?:been|[dg]one)",
        apply=lambda s: s.replace("'s", " has"),
    ),
    dict(regexp=r"(?:there|s?he|it)'s", apply=lambda s: s.replace("'s", " is")),
    dict(
        regexp=r"\s(?:[Aa]pi|[Ii]p|[Ii]d|[Uu]r[il])",
        plural=r"s?[\s\.,]",
        apply=str.upper,
    ),
    dict(regexp=r"\s(?:[Jj]son|[Dd]ns|[Hh]tml|[Vv]m)[\s\.,]", apply=str.upper),
]

OFFENDING_SPEC = [
    (re.compile(f"(.*[^(])({x['regexp']})({x.get('plural', '')})([^)]?.*)"), x)
    for x in OFFENDING_SPEC
]


class YAMLDocException(Exception):
    def __init__(self, *args: object) -> None:
        super().__init__()
        self.args = args


class AnsibleDocProcessor:
    def __init__(
        self,
        indent: int = 2,
        width: int = 120,
        offenders: bool = False,
        fix_offenders: bool = False,
        dry_run: bool = False,
    ):
        self.indent = indent
        self.width = width
        self.offenders = offenders
        self.fix_offenders = fix_offenders
        self.dry_run = dry_run
        self.yaml_indents = self._calculate_indent(indent)
        self.yaml = self.make_yaml_instance()
        self.first_line_no = 0
        self.json_samples = {}
        self.json_sample_id_count = 0

    @staticmethod
    def _calculate_indent(num: int) -> Dict[str, int]:
        return dict(mapping=num, sequence=num + 2, offset=2)

    def make_yaml_instance(self) -> YAML:
        if IMPORT_ERROR:
            raise AndeboxException(
                "Missing dependency for action 'yaml-doc': "
            ) from IMPORT_ERROR

        yaml = YAML()
        yaml.indent(**self.yaml_indents)
        yaml.width = self.width
        yaml.preserve_quotes = True
        yaml.explicit_start = False
        yaml.preserve_quotes = True
        yaml.top_level_colon_align = False
        yaml.compact_seq_seq = False
        return yaml

    def read_yaml(self, content: str) -> Optional[Union[Dict[str, Any], list]]:
        return self.yaml.load(content)

    def dump_yaml(self, data: Union[Dict[str, Any], list]) -> str:
        output = StringIO()
        self.yaml.dump(data, output)
        return output.getvalue()

    def fix_desc_str(self, line: str) -> str:
        line = line.strip()
        line = re.sub(r"\s\s+", " ", line)
        if not line[0].isupper():
            line = line[0].upper() + line[1:]
        if line.endswith(".)"):
            return f"{line[:-2]})."
        if line.endswith(DESCRIPTION_ACCEPTED_END_CHARS):
            return line
        return f"{line}."

    def process_description(self, desc: Union[List[str], str]) -> Union[List[str], str]:
        try:
            if isinstance(desc, str):
                return self.fix_desc_str(desc)
            else:
                return [self.fix_desc_str(x) for x in desc]
        except Exception as e:
            raise YAMLDocException(desc) from e

    def _store_json_sample(self, sample: Any) -> str:
        json_str = json.dumps(sample, indent=self.indent)
        sample_hash = hashlib.md5(
            (json_str + str(self.json_sample_id_count)).encode()
        ).hexdigest()[:8]
        sample_id = f"{JSON_SAMPLE_PREFIX}-{sample_hash}"
        self.json_samples[sample_id] = json_str
        self.json_sample_id_count += 1
        return sample_id

    def process_sample(self, sample: Any, type_: str) -> Any:
        output_sample = self.dump_yaml(sample)
        is_json = (type_ == "list" and output_sample.strip().startswith("[")) or (
            type_ == "dict" and output_sample.strip().startswith("{")
        )
        if not is_json:
            return sample
        # Store the JSON and return a placeholder ID
        return self._store_json_sample(sample)

    def process_options(
        self, opts: Dict[str, Any], suboptions_kw: str
    ) -> Dict[str, Any]:
        try:
            for option in opts.values():
                if "description" in option:
                    option["description"] = self.process_description(
                        option["description"]
                    )
                if suboptions_kw in option:
                    option[suboptions_kw] = self.process_options(
                        option[suboptions_kw], suboptions_kw
                    )
                if "sample" in option and option["type"] in ("list", "dict"):
                    option["sample"] = self.process_sample(
                        option["sample"], option["type"]
                    )
            return opts
        except Exception as e:
            raise YAMLDocException(opts, suboptions_kw) from e

    def process_documentation(self, data: Dict[str, Any]) -> Dict[str, Any]:
        try:
            if data.get("short_description", " ").endswith("."):
                data["short_description"] = data["short_description"].rstrip(".")
            if desc := data.get("description"):
                data["description"] = self.process_description(desc)
            if notes := data.get("notes"):
                data["notes"] = self.process_description(notes)
            if seealso := data.get("seealso"):
                for sa in seealso:
                    if sa_desc := sa.get("description"):
                        sa["description"] = self.process_description(sa_desc)
            if options := data.get("options"):
                data["options"] = self.process_options(options, "suboptions")
            return data
        except Exception as e:
            raise YAMLDocException(data) from e

    def process_return(self, data: Dict[str, Dict]) -> Dict[str, Dict]:
        try:
            data = self.process_options(data, "contains")
            return data
        except Exception as e:
            raise YAMLDocException(data) from e

    def get_processor(self, variable: str, in_doc_fragments: bool = False) -> Callable:
        processors_map = {
            "DOCUMENTATION": self.process_documentation,
            "RETURN": self.process_return,
        }
        return processors_map.get(
            variable, self.process_documentation if in_doc_fragments else lambda x: x
        )

    def process_yaml(self, quoted_content: List[str], processor: Callable) -> List[str]:
        data = self.read_yaml("\n".join(quoted_content))
        if not data:
            return quoted_content

        data = processor(data)
        yaml_content = self.dump_yaml(data).splitlines()

        if isinstance(data, list):
            yaml_content = [
                (line[2:] if line.startswith("  ") else line) for line in yaml_content
            ]
        while yaml_content and yaml_content[0] == "":
            yaml_content.pop(0)

        return quoted_content if self.dry_run else yaml_content

    def postprocess_content(self, in_variable: str, content: List[str]) -> List[str]:
        content = self.postprocess_json_samples(content)

        if self.offenders and in_variable != "EXAMPLES":
            fixed_content = self.process_offenders(content)
            if self.fix_offenders:
                content = fixed_content
        if in_variable != "EXAMPLES":
            fixed_content = self.postprocess_line_length(content)
            if self.fix_offenders:
                content = fixed_content
        return content

    def apply_offender_rule(self, line_num: int, line: str) -> str:
        for regexp, spec in OFFENDING_SPEC:
            if match := regexp.match(line):
                if self.offenders and not self.fix_offenders:
                    print(f"  {line_num:4}: {line}")
                prefix, term, plural, suffix = match.groups()

                if func := spec.get("apply"):
                    return f"{prefix}{func(term)}{plural}{suffix}"
                elif repl := spec.get("replace"):
                    return f"{prefix}{repl}{plural}{suffix}"

        return line

    def process_offenders(self, content: List[str]) -> List[str]:
        result = []
        for num, line in enumerate(content):
            line_num = 2 + self.first_line_no + num

            def apply(line):
                return self.apply_offender_rule(line_num, line)

            prev_line = fixed_line = line
            while (fixed_line := apply(fixed_line)) != prev_line:
                prev_line = fixed_line

            if fixed_line != line:
                print(f"  {line_num:4}: {fixed_line}")
            result.append(fixed_line)

        return result

    def postprocess_json_samples(self, content: List[str]) -> List[str]:
        result = []
        sample_pattern = re.compile(
            rf"^(\s+sample:)\s+({JSON_SAMPLE_PREFIX}-[0-9a-f]{{8}})$"
        )

        for line in content:
            match = sample_pattern.match(line)
            if (not match) or match.group(2) not in self.json_samples:
                result.append(line)
                continue

            indentend_sample_key = match.group(1)
            sample_id = match.group(2)

            json_lines = self.json_samples[sample_id].splitlines()
            base_indent = " " * (
                len(indentend_sample_key) - len("sample:") + self.indent
            )
            result.append(f"{indentend_sample_key}")
            result.extend(f"{base_indent}{line}" for line in json_lines)

            del self.json_samples[sample_id]

        return result

    def postprocess_line_length(self, content: List[str]) -> List[str]:
        """Post-process line lengths."""
        hard_limit = 160
        LINE_RE = re.compile(r"^(\s*[-\s]\s)\S.*")
        results = []

        for line in content:
            if len(line) <= hard_limit:
                results.append(line)
                continue

            if match := LINE_RE.match(line):
                lead_spaces = len(match.group(1)) * " "
                line_split = line.split(" ")
                first_part = f'{" ".join(line_split[:-1])}'
                last_part = f"{lead_spaces}{line_split[-1]}"

                if len(first_part) > hard_limit:
                    first_part = f"{first_part} {FIXME_TAG}"
                if len(last_part) > hard_limit:
                    last_part = f"{last_part} {FIXME_TAG}"
                results.append(first_part)
                results.append(last_part)
            else:
                results.append(f"{line} {FIXME_TAG}")

        return results

    def process_file(self, file_path: Path) -> None:
        """Process a single file."""
        QUOTE_RE_FRAG = r'(?:"|\'){3}'
        VAR_RE_FRAG = r"(?:[A-Z]+)"
        ONELINE_RE = re.compile(
            rf"^\s*{VAR_RE_FRAG}\s*=\s*r?{QUOTE_RE_FRAG}.*{QUOTE_RE_FRAG}$"
        )
        FIRST_LINE_RE = re.compile(rf"^(\s*{VAR_RE_FRAG})\s*=\s*r?{QUOTE_RE_FRAG}(.*)$")

        print(f"Opening file {file_path}")
        with open(file_path, "r") as file:
            lines = file.readlines()

        updated_lines = []
        is_doc_frag = "doc_fragments" in file_path.parts
        in_variable = ""
        first_line = ""
        quoted_content = []
        self.first_line_no = 0

        for line_no, line in enumerate(lines):
            line = line.rstrip()
            if in_variable:
                if not re.match(rf"^\s*{QUOTE_RE_FRAG}", line):
                    quoted_content.append(line)
                    continue

                updated_lines.append(first_line)
                processor = self.get_processor(
                    "DOCUMENTATION" if is_doc_frag else in_variable
                )
                outbound_content = self.process_yaml(quoted_content, processor)
                updated_lines.extend(
                    self.postprocess_content(in_variable, outbound_content)
                )
                updated_lines.append('"""')

                in_variable = ""
                quoted_content = []
                self.first_line_no = 0
            else:
                if ONELINE_RE.search(line):
                    updated_lines.append(re.sub(QUOTE_RE_FRAG, '"""', line))
                elif match := FIRST_LINE_RE.search(line):
                    in_variable, yaml_first_line = match.groups()
                    if yaml_first_line.endswith("---"):
                        yaml_first_line = yaml_first_line[:-3]
                    first_line = f'{in_variable} = r"""{yaml_first_line}'
                    self.first_line_no = line_no
                else:
                    updated_lines.append(line)

        if not self.dry_run:
            with open(file_path, "w") as file:
                file.writelines([f"{x}\n" for x in updated_lines])


class YAMLDocAction(AndeboxAction):
    name = "yaml-doc"
    help = "analyze and/or reformat YAML documentation in plugins"
    args = [
        dict(
            names=("--offenders", "-o"),
            specs=dict(
                action="store_true",
                help="report potential style-related offending constructs",
            ),
        ),
        dict(
            names=("--fix-offenders", "-O"),
            specs=dict(
                action="store_true",
                help="fix potential style-related offending constructs, implies (--offenders)",
            ),
        ),
        dict(
            names=("--dry_run", "-n"),
            specs=dict(
                action="store_true",
                help="do not modify files",
            ),
        ),
        dict(
            names=("--width", "-w"),
            specs=dict(
                type=int,
                default=120,
                help="width for the YAML output (default: 120)",
            ),
        ),
        dict(
            names=("--indent", "-i"),
            specs=dict(
                type=int,
                default=2,
                help="indentation for the YAML output (default: 2)",
            ),
        ),
        dict(
            names=("files",),
            specs=dict(
                help="Files where to search for YAML content",
                type=Path,
                nargs="+",
            ),
        ),
    ]

    def run(self, context):
        processor = AnsibleDocProcessor(
            indent=context.args.indent,
            width=context.args.width,
            offenders=context.args.offenders or context.args.fix_offenders,
            fix_offenders=context.args.fix_offenders,
            dry_run=context.args.dry_run,
        )

        for file_path in context.args.files:
            processor.process_file(file_path)


# code: language=python tabSize=4



# -*- coding: utf-8 -*-
# Copyright (c) 2021, Alexei Znamensky
# All rights reserved.
#
# This file is part of the Andebox project and is distributed under the terms
# of the BSD 3-Clause License. See LICENSE file for details.


class AndeboxException(Exception):
    pass


# code: language=python tabSize=4



# -*- coding: utf-8 -*-
# Copyright (c) 2021, Alexei Znamensky
# All rights reserved.
#
# This file is part of the Andebox project and is distributed under the terms
# of the BSD 3-Clause License. See LICENSE file for details.
import os
import re
import sys
from dataclasses import dataclass
from functools import reduce
from functools import total_ordering
from pathlib import Path

from looseversion import LooseVersion

from .base import AndeboxAction


class IgnoreFileEntry:
    pattern = re.compile(
        r"^(?P<filename>\S+)\s(?P<ignore>\S+)(?:\s+#\s*(?P<comment>\S.*\S))?\s*$"
    )
    filter_files = None
    filter_checks = None
    file_parts_depth = None

    def __init__(self, filename, ignore, comment):
        self.filename = filename
        self._file_parts = self.filename.split("/")

        if ":" in ignore:
            self.ignore, self.error_code = ignore.split(":")
        else:
            self.ignore, self.error_code = ignore, None
        self.comment = comment

    @property
    def ignore_check(self):
        return f"{self.ignore}:{self.error_code}" if self.error_code else self.ignore

    @property
    def rebuilt_comment(self):
        return f" # {self.comment}" if self.comment else ""

    @property
    def file_parts(self):
        if self.file_parts_depth is None:
            return str(Path(*self._file_parts))

        return str(Path(*self._file_parts[: self.file_parts_depth]))

    def __str__(self):
        return f"<IgnoreFileEntry: {self.filename} {self.ignore_check}{self.rebuilt_comment}>"

    def __repr__(self):
        return str(self)

    @staticmethod
    def parse(line):
        match = IgnoreFileEntry.pattern.match(line)
        if not match:
            raise ValueError(f"Line cannot be parsed as an ignore-file entry: {line}")

        ffilter = IgnoreFileEntry.filter_files
        if ffilter is not None:
            ffilter = (
                ffilter if isinstance(ffilter, re.Pattern) else re.compile(ffilter)
            )
            if not ffilter.search(match.group("filename")):
                return None

        ifilter = IgnoreFileEntry.filter_checks
        if ifilter is not None:
            ifilter = (
                ifilter if isinstance(ifilter, re.Pattern) else re.compile(ifilter)
            )
            if not ifilter.search(match.group("ignore")):
                return None

        return IgnoreFileEntry(
            match.group("filename"), match.group("ignore"), match.group("comment")
        )


# pragma: no cover
@total_ordering
@dataclass
class ResultLine:
    file_part: str
    ignore_check: str
    count: int = 1

    def increase(self) -> "ResultLine":
        self.count = self.count + 1
        return self

    def __lt__(self, other) -> bool:
        return self.count < other.count

    def __str__(self) -> str:
        r = [f"{self.count:6} "]
        if self.file_part:
            r.append(" ")
            r.append(self.file_part)
        if self.ignore_check:
            r.append(" ")
            r.append(self.ignore_check)
        return "".join(r)

    def __repr__(self) -> str:
        r = ["<ResultLine: ", str(self.count), ","]
        if self.file_part:
            r.append(" ")
            r.append(self.file_part)
        if self.ignore_check:
            r.append(" ")
            r.append(self.ignore_check)
        r.append(">")
        return "".join(r)


# @TODO works only for collections because of the hardcoded path
_ignore_path = Path(".") / "tests" / "sanity"
try:
    with os.scandir(_ignore_path) as sanity_dir:
        _ignore_versions = sorted(
            [
                str(LooseVersion(entry.name[7:-4]))
                for entry in sanity_dir
                if entry.name.startswith("ignore-") and entry.name.endswith(".txt")
            ]
        )
except FileNotFoundError:
    _ignore_versions = []


class IgnoreLinesAction(AndeboxAction):
    name = "ignores"
    help = "gathers stats on ignore*.txt file(s)"
    args = [
        dict(
            names=["--ignore-file-spec", "-ifs"],
            specs=dict(
                choices=_ignore_versions + ["-"],
                help=(
                    "use the ignore file matching this Ansible version; "
                    "special value '-' may be specified to read "
                    "from stdin instead; if not specified, will use all available files; "
                    "if no choices are presented, the collection structure was not recognized"
                ),
            ),
        ),
        dict(
            names=["--depth", "-d"],
            specs=dict(type=int, help="path depth for grouping files"),
        ),
        dict(
            names=["--filter-files", "-ff"],
            specs=dict(
                type=re.compile,
                help="regular expression matching file names to be included",
            ),
        ),
        dict(
            names=["--filter-checks", "-fc"],
            specs=dict(
                type=re.compile,
                help="regular expression matching checks in ignore files to be included",
            ),
        ),
        dict(
            names=["--suppress-files", "-sf"],
            specs=dict(
                action="store_true",
                help="supress file names from the output, consolidating the results",
            ),
        ),
        dict(
            names=["--suppress-checks", "-sc"],
            specs=dict(
                action="store_true",
                help="suppress the checks from the output, consolidating the results",
            ),
        ),
        dict(
            names=["--head", "-H"],
            specs=dict(
                type=int,
                default=10,
                help=(
                    "number of lines to display in the output: leading lines if "
                    "positive, trailing lines if negative, all lines if zero."
                ),
            ),
        ),
    ]

    # pylint: disable=consider-using-with
    def make_fh_list_for_version(self, version):
        if version == "-":
            return [sys.stdin]
        if version:
            return [open(_ignore_path / f"ignore-{version}.txt")]

        with os.scandir(_ignore_path) as it:
            return [
                open(_ignore_path / entry.name)
                for entry in it
                if entry.name.startswith("ignore-") and entry.name.endswith(".txt")
            ]

    @staticmethod
    def read_ignore_file(fh):
        result = []
        with fh:
            for line in fh.readlines():
                entry = IgnoreFileEntry.parse(line)
                if entry:
                    result.append(entry)
        return result

    def retrieve_ignore_entries(self, version):
        return reduce(
            lambda a, b: a + b,
            [
                self.read_ignore_file(fh)
                for fh in self.make_fh_list_for_version(version)
            ],
        )

    @staticmethod
    def filter_lines(lines, num):
        if num == 0:
            return lines
        return lines[num:] if num < 0 else lines[:num]

    def run(self, context):
        if context.args.filter_files:
            IgnoreFileEntry.filter_files = context.args.filter_files
        if context.args.filter_checks:
            IgnoreFileEntry.filter_checks = context.args.filter_checks
        if context.args.depth:
            IgnoreFileEntry.file_parts_depth = context.args.depth

        try:
            ignore_entries = self.retrieve_ignore_entries(context.args.ignore_file_spec)
        except Exception as e:
            print(
                f"Error reading ignore file {context.args.ignore_file_spec}: {e}",
                file=sys.stderr,
            )
            raise e

        count_map = {}
        for entry in ignore_entries:
            fp = entry.file_parts if not context.args.suppress_files else ""
            ic = entry.ignore_check if not context.args.suppress_checks else ""
            key = fp + "|" + ic
            count_map[key] = count_map.get(key, ResultLine(fp, ic, 0)).increase()

        lines = [str(s) for s in sorted(count_map.values(), reverse=True)]
        print("\n".join(self.filter_lines(lines, context.args.head)))


# code: language=python tabSize=4



# -*- coding: utf-8 -*-
# Copyright (c) 2021-2023, Alexei Znamensky
# All rights reserved.
#
# This file is part of the Andebox project and is distributed under the terms
# of the BSD 3-Clause License. See LICENSE file for details.
import os
import subprocess
import sys

from ..exceptions import AndeboxException
from .base import AndeboxAction


def _make_default_tox_ini():
    return f""";
; andebox tox-test's tox.ini -- this file is not overwritten by andebox
[tox]
isolated_build = true
envlist = ac215, ac216, ac217, ac218, dev
skipsdist = true

[testenv]
passenv = PWD, HOME
skip_install = true
allowlist_externals = andebox
basepython = {sys.executable}
deps =
  andebox>=0.36
  ac215: ansible-core~=2.15.0
  ac216: ansible-core~=2.16.0
  ac217: ansible-core~=2.17.0
  ac218: ansible-core~=2.18.0
  dev: https://github.com/ansible/ansible/archive/devel.tar.gz
commands = andebox test -- {{posargs}}
"""


class ToxTestError(AndeboxException):
    pass


class ToxTestAction(AndeboxAction):
    name = "tox-test"
    help = "runs ansible-test within tox, for testing in multiple ansible versions"
    args = [
        dict(
            names=("--env", "-e"),
            specs=dict(help="tox environments to run the test in"),
        ),
        dict(
            names=("--list", "-l"),
            specs=dict(action="store_true", help="list all tox environments (tox -a)"),
        ),
        dict(
            names=("--recreate", "-r"),
            specs=dict(
                action="store_true",
                help="force recreation of virtual environments (tox -r)",
            ),
        ),
        dict(names=("ansible_test_params",), specs=dict(nargs="*")),
    ]
    tox_ini_filename = ".andebox-tox-test.ini"

    @classmethod
    def make_parser(cls, subparser):
        action_parser = super(ToxTestAction, cls).make_parser(subparser)
        action_parser.epilog = (
            "Notice the use of '--' to delimit andebox's options from tox's"
        )
        action_parser.usage = "%(prog)s [-h] [--env ENV] -- [ansible_test_params ...]"

    def run(self, context):
        if not os.path.exists(self.tox_ini_filename):
            with open(self.tox_ini_filename, "w") as tox_ini:
                tox_ini.write(_make_default_tox_ini())

        cmd_args = ["tox", "-c", self.tox_ini_filename]
        if context.args.list:
            cmd_args.append("-a")
        if context.args.recreate:
            cmd_args.append("-r")
        if context.args.env:
            cmd_args.extend(["-e", context.args.env])
        cmd_args.append("--")
        cmd_args.extend(context.args.ansible_test_params)
        rc = subprocess.call(cmd_args)

        if rc != 0:
            raise ToxTestError(f"Error running tox (rc={rc})")


# code: language=python tabSize=4



# -*- coding: utf-8 -*-
# (c) 2021, Alexei Znamensky <russoz@gmail.com>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
from ..context import ConcreteContext
from ..context import ContextType
from .base import AndeboxAction


def printline(title, content):
    print(f"{title:>20}: {content}")


class ContextAction(AndeboxAction):
    name = "context"
    help = "returns information from running context"
    args = []

    def run(self, context: ConcreteContext):
        printline("Base dir", context.base_dir)
        if context.venv:
            printline("Venv", context.venv)
        printline("Type", context.type)
        printline("Temp dir", context.top_dir)
        printline("ansible-test", context.ansible_test)
        printline("Sanity tests", context.sanity_test_subdir)
        printline("Integration tests", context.integration_test_subdir)

        if context.type == ContextType.COLLECTION:
            ns, name, version = context.read_coll_meta()  # type: ignore
            printline("Collection", f"{ns}.{name} {version}")


# code: language=python tabSize=4



# -*- coding: utf-8 -*-
# Copyright (c) 2021-2023, Alexei Znamensky
# All rights reserved.
#
# This file is part of the Andebox project and is distributed under the terms
# of the BSD 3-Clause License. See LICENSE file for details.
import os
from contextlib import contextmanager
from pathlib import Path
from typing import Any
from typing import Generator


@contextmanager
def set_dir(path: Path) -> Generator[Any, Any, Any]:
    previous = Path().absolute()
    try:
        os.chdir(path)
        yield
    finally:
        os.chdir(previous)


# code: language=python tabSize=4



# -*- coding: utf-8 -*-
# Copyright (c) 2023, Alexei Znamensky
# All rights reserved.
#
# This file is part of the Andebox project and is distributed under the terms
# of the BSD 3-Clause License. See LICENSE file for details.
import os
import shutil
import subprocess
import sys
import tempfile
import time
from abc import ABC
from abc import abstractmethod
from argparse import Namespace
from contextlib import contextmanager
from enum import Enum
from pathlib import Path
from typing import Any
from typing import Generator
from typing import Tuple
from typing import Type

import yaml

from .exceptions import AndeboxException

toplevel_exclusion = (".git", ".tox", ".venv", ".virtualvenv", "venv", "virtualenv")


class AndeboxUnknownContext(AndeboxException):
    pass


class ContextType(Enum):
    ANSIBLE_CORE = 1
    COLLECTION = 2


class AbstractContext(ABC):
    _context_type: ContextType = None  # type: ignore

    def __init__(self, base_dir: Path, args: Namespace) -> None:
        self.base_dir = base_dir
        self.args = args
        self.venv = args.venv
        self.top_dir = Path(tempfile.mkdtemp(prefix="andebox."))

    @property
    def type(self) -> ContextType:
        return self._context_type  # type: ignore

    @property
    @abstractmethod
    def ansible_test(self) -> str:
        pass

    @property
    @abstractmethod
    def sub_dir(self) -> str:
        pass

    @property
    def full_dir(self) -> Path:
        return self.top_dir / self.sub_dir

    @property
    def tests_subdir(self) -> Path:
        return Path("tests")

    @property
    def sanity_test_subdir(self) -> Path:
        return self.tests_subdir / "sanity"

    @property
    def unit_test_subdir(self) -> Path:
        return self.tests_subdir / "unit"

    @property
    def integration_test_subdir(self) -> Path:
        return self.tests_subdir / "integration"

    def install_requirements(self, reqs: Path, path: Path | None, retries: int) -> None:
        pass

    @abstractmethod
    def post_sub_dir(self, top_dir: Path) -> None:
        pass

    def copy_tree(self) -> None:
        # copy files to tmp ansible coll dir
        with os.scandir() as it:
            for entry in it:
                if any(entry.name.startswith(x) for x in toplevel_exclusion):
                    continue
                if entry.is_dir():
                    shutil.copytree(
                        entry.name,
                        self.full_dir / entry.name,
                        symlinks=True,
                        ignore_dangling_symlinks=True,
                    )
                else:
                    shutil.copy(
                        entry.name,
                        self.full_dir / entry.name,
                        follow_symlinks=False,
                    )

    @contextmanager
    def temp_tree(self) -> Generator[Path, Any, Any]:
        self.full_dir.mkdir(parents=True, exist_ok=True)
        print(f"directory  = {self.full_dir}", file=sys.stderr)
        self.copy_tree()

        self.post_sub_dir(self.top_dir)

        yield self.full_dir

        if self.args.keep:
            print(f"Keeping temporary directory: {self.full_dir}")
        else:
            print(f"Removing temporary directory: {self.full_dir}")
            shutil.rmtree(self.top_dir)

    def copy_exclude_lines(
        self, src: Path, dest: Path, exclusion_filenames: list[str]
    ) -> None:
        with src.open("r") as src_file, dest.open("w") as dest_file:
            for line in src_file.readlines():
                if not any(line.startswith(f) for f in exclusion_filenames):
                    dest_file.write(line)

    def binary_path(self, binary: str) -> str:
        if self.args.venv:
            return str(Path(self.args.venv) / "bin" / binary)

        return str(Path(binary))

    def exclude_from_ignore(self) -> None:
        files = [f for f in self.args.ansible_test_params if Path(f).is_file()]
        print(f"Excluding from ignore files: {files}")
        if self.args.exclude_from_ignore:
            src_dir = Path.cwd() / self.sanity_test_subdir
            dest_dir = self.full_dir / self.sanity_test_subdir
            with os.scandir(src_dir) as ts_dir:
                for ts_entry in ts_dir:
                    if ts_entry.name.startswith("ignore") and ts_entry.name.endswith(
                        ".txt"
                    ):
                        self.copy_exclude_lines(
                            src_dir / ts_entry.name,
                            dest_dir / ts_entry.name,
                            files,
                        )


class AnsibleCoreContext(AbstractContext):
    _context_type = ContextType.ANSIBLE_CORE

    @property
    def ansible_test(self) -> str:
        return str(self.top_dir / "bin" / "ansible-test")

    @property
    def sub_dir(self) -> str:
        return ""

    def post_sub_dir(self, top_dir: Path) -> None:
        pass

    @property
    def tests_subdir(self) -> Path:
        return Path("test")

    @property
    def unit_test_subdir(self) -> Path:
        return self.tests_subdir / "units"


class CollectionContext(AbstractContext):
    _context_type = ContextType.COLLECTION

    def __init__(self, base_dir: Path, args: Namespace) -> None:
        super().__init__(base_dir, args)
        self.name = self.version = ""
        self.namespace, self.collection = self.determine_collection(
            self.args.collection
        )

    @property
    def ansible_test(self) -> str:
        return self.binary_path("ansible-test")

    @property
    def sub_dir(self) -> Path:
        coll_dir = Path("ansible_collections") / self.namespace / self.collection
        return coll_dir

    def post_sub_dir(self, top_dir: Path) -> None:
        print(f"collection = {self.namespace}.{self.collection}", file=sys.stderr)
        os.putenv(
            "ANSIBLE_COLLECTIONS_PATH",
            ":".join(
                [str(top_dir)]
                + os.environ.get("ANSIBLE_COLLECTIONS_PATH", "").split(":")
            ),
        )

    def read_coll_meta(self) -> tuple[str, str, str]:
        with open("galaxy.yml") as galaxy_meta:
            meta = yaml.safe_load(galaxy_meta)
        self.namespace, self.name, self.version = (
            meta["namespace"],
            meta["name"],
            meta["version"],
        )
        return meta["namespace"], meta["name"], meta["version"]

    def determine_collection(self, coll_arg: str) -> tuple[str, str]:
        if coll_arg:
            coll_split = coll_arg.split(".")
            return ".".join(coll_split[:-1]), coll_split[-1]
        return self.read_coll_meta()[:2]

    def install_requirements(self, reqs: Path, path: Path | None, retries: int) -> None:
        # reqs = self.integration_test_subdir / "requirements.yml"
        if not reqs.exists():
            print(f"Cannot find requirements file: {reqs}")
            return

        print(f"Installing requirements ({reqs})", end="")
        if path:
            print(f" at path: {path}")
        else:
            print("")
        path_arg = ["-p", f"{path}"] if path else []

        cmd = (
            [self.binary_path("ansible-galaxy"), "collection", "install"]
            + path_arg
            + ["-r", f"{reqs}", "-vvv", "--force"]
        )
        print(f"Running: {cmd}")
        delay = 10
        for attempt in range(1, retries + 1):
            try:
                subprocess.run(cmd, check=True)
                break
            except subprocess.CalledProcessError:
                if attempt == retries:
                    raise
                print(
                    f"Install requirements failed (attempt {attempt}/{retries}), retrying in {delay}s..."
                )
                time.sleep(delay)


ConcreteContextType = Type[AnsibleCoreContext] | Type[CollectionContext]
ConcreteContext = AnsibleCoreContext | CollectionContext


def _base_dir_type(
    dir_: Path,
) -> ConcreteContextType:
    if (dir_ / "bin" / "ansible-playbook").exists():
        return AnsibleCoreContext
    if (dir_ / "meta" / "runtime.yml").exists():
        return CollectionContext
    raise ValueError()


def _determine_base_dir_rec(
    dir_: Path,
) -> Tuple[Path, ConcreteContextType]:
    try:
        return dir_, _base_dir_type(dir_)
    except ValueError:
        if dir_ == Path(dir_.anchor):
            raise AndeboxUnknownContext()  # pylint: disable=raise-missing-from
        return _determine_base_dir_rec(dir_.parent)


def _determine_base_dir() -> Tuple[Path, ConcreteContextType]:
    cur_dir = Path.cwd()
    try:
        return _determine_base_dir_rec(cur_dir)
    except AndeboxUnknownContext as e:
        raise AndeboxUnknownContext(
            f"Cannot determine current directory context: f{cur_dir}"
        ) from e


def create_context(args: Namespace) -> ConcreteContext:
    base_dir, basedir_type = _determine_base_dir()
    return basedir_type(base_dir, args)


# code: language=python tabSize=4



# -*- coding: utf-8 -*-
# Copyright (c) 2024, Alexei Znamensky
# All rights reserved.
#
# This file is part of the Andebox project and is distributed under the terms
# of the BSD 3-Clause License. See LICENSE file for details.
import os
import subprocess
import webbrowser
from pathlib import Path

from ..context import CollectionContext
from ..context import ContextType
from ..exceptions import AndeboxException
from ..util import set_dir
from .base import AndeboxAction


class DocsiteAction(AndeboxAction):
    name = "docsite"
    help = "builds collection docsite"
    args = [
        dict(
            names=("--keep", "-k"),
            specs=dict(
                help="keep temporary collection directory after execution",
                action="store_true",
            ),
        ),
        dict(
            names=("--open", "-o"),
            specs=dict(
                help="open browser pointing to main page after build",
                action="store_true",
            ),
        ),
        dict(
            names=("--dest-dir", "-d"),
            specs=dict(
                help="directory which should contain the docsite",
                default=".builtdocs",
                type=Path,
            ),
        ),
    ]

    def run(self, context: CollectionContext):
        if context.type != ContextType.COLLECTION:
            raise AndeboxException(
                "Action 'docsite' must be executed in a collection context!"
            )
        try:
            with context.temp_tree() as collection_dir:
                os.makedirs(context.args.dest_dir, mode=0o755, exist_ok=True)
                if not (context.args.dest_dir / "build.sh").exists():
                    subprocess.run(
                        [
                            context.binary_path("antsibull-docs"),
                            "sphinx-init",
                            "--use-current",
                            "--lenient",
                            f"{context.namespace}.{context.name}",
                            "--dest-dir",
                            f"{context.args.dest_dir}",
                        ],
                        cwd=collection_dir,
                        check=True,
                    )

                with set_dir(context.args.dest_dir):
                    subprocess.run(
                        [
                            context.binary_path("python"),
                            "-m",
                            "pip",
                            "install",
                            "-qr",
                            "requirements.txt",
                        ],
                        check=True,
                    )
                    subprocess.run(["./build.sh"], check=True)
        except Exception as e:
            raise AndeboxException(f"Error running when building docsite: {e}") from e

        if context.args.open:
            webbrowser.open(
                f"{context.args.dest_dir / 'build' / 'html' / 'index.html'}"
            )


# code: language=python tabSize=4
